#file with functions providing FFT homogenization of linear elasticity
from numpy import *
from numpy.fft import *
from matplotlib import *
from scipy import *
import sys
import dbg

def homogenization(inputs):
    """FFTH homogenization of linear elasticity
    INPUT = 
        inputs : dictionary consisting of following variables:
            L : ndarray of stiffness coefficients
            K0 : bulk modulus of reference stiffness
            G0 : shear modulus of reference stiffness
            Y : size of periodic unit cell
            tol : tolerance required for convergence
            maxit : maximal number of iterations
    OUTPUT = 
        L_eff : homogenized (effective) stiffness
        kit : number of iterations reached at convergence
        resnorm : residual based norm reached during iterations
        time : the total time reached during each loading
        L : stiffness
        strain : strain of individual loading cases
        Y : the size of periodic unit cell
    """
    d,D,N,DN = get_dimensions(inputs['L'][0][0])
#    L0 = get_stiff_Man(inputs['K0'], inputs['G0'])
    # calculation
    Gam = get_Gamma(N, d, D, inputs['Y'], inputs['K0'], inputs['G0'])
    L_eff = zeros([D,D])#global macroscale stiffness - preallocation
    strain_data = zeros(hstack([D,D,N]))
    kit = zeros(D); resnorm = ones(D); time = zeros(D)
    for iL in arange(D):
        t = dbg.start_time() 
        ep0 = zeros(D)
        ep0[iL] = 1
        X = zeros(DN)
        B = zeros(DN) 
        Bep0 = zeros(DN)
        for i in arange(D):
            Bep0[i] = ep0[i]
        B = - mult_PL(Bep0,inputs['L'],Gam,D,DN)
        X, resnorm[iL], kit[iL] = CG_solver(inputs['L'],B,X,Gam,inputs['tol'],inputs['maxit'])
        strain = X + Bep0
        stress = mult_Lep(inputs['L'],strain,D,DN)
#        eqnorm = equinorm(stress,inputs['Y'],N,d)            
        for jL in arange(D):
            L_eff[jL][iL] = mean(stress[jL])
        time[iL] = dbg.get_time(t)
        strain_data[iL] = strain
    return dict(L_eff=L_eff,kit=kit,resnorm=resnorm,time=time,L=inputs['L'],strain=strain_data,Y=inputs['Y'])

def CG_solver(L,B,X,Gam,tol,maxit):
    """Solve the linear system with stiffness matrix L and right hand side B; 
        it is useful for both linear elasticity and visco-elasticity 
    INPUT = 
        L : ndarray of stiffness coefficients
        tol : tolerance required for convergence
        maxit : maximal number of iterations
        B : ndarray of right hand side
        Gam : ndarray of Kernel operator
        X : initial vector
    OUTPUT = 
        X : the solution of linear system with zero mean
        resnorm : the norm of residual
        kit : number of iterations
    """
    dummy,D,dummy,DN = get_dimensions(L[0][0])
    AX = mult_PL(X,L,Gam,D,DN)
    R = B - AX
    P = R
    rr = dot(R.flatten(1),R.flatten(1))
    kit = 0; resnorm = sqrt(rr)
    while (resnorm > tol) == (kit < maxit):
        kit = kit + 1
        AP = mult_PL(P,L,Gam,D,DN)
        alp = rr/dot(P.flatten(1),AP.flatten(1))
        X = X + alp*P
        R = R - alp*AP
        rrnext = dot(R.flatten(1),R.flatten(1))
        bet = rrnext/rr
        rr = rrnext
        P = R+bet*P
        resnorm = sqrt(rr)
    return X, resnorm, kit

def get_dimensions(data):
    """Solve the linear system with stiffness matrix L and right hand side B; 
        it is useful for both linear elasticity and visco-elasticity 
    INPUT = 
        data : ndarray of e.g. stiffness coefficients
    OUTPUT = 
        d : dimensionality of the problem; in our case of the two-dimensional problem d = 2
        D : the dimensionality of linear elasticity in engineering notation; D = 3
        N : the number of discretization points; a vector of size d
        DN : the vector consisting the number of components in the dimension D; useful for storing strain or stress components at nodal points
    """
    d = ndim(data)
    D = d*(d+1)/2
    N = array(shape(data))
    DN = hstack([D,N])
    return d,D,N,DN


def get_Gamma(N,d,D,Y,K0,G0):
    """Projection matrix on a space of admissible strain fields 
    INPUT = 
        N : ndarray of e.g. stiffness coefficients
        d : dimension; d = 2 
        D : dimension in engineering notation; D = 3
        Y : the size of periodic unit cell
        K0 : reference bulk modulus, parameter
        G0 : reference strain modulus, parameter
    OUTPUT = 
        Gamma : projection matrix of size DxDxN, i.e. as stiffness matrix L
    """   
    xi = get_xi(Y,N,d)
    xi2 = []
    for i in arange(d):
        xi2.append(xi[i]**2)
    C = [1./(4*G0),(3.*K0+G0)/G0/(3*K0+4*G0)]
    Gamma = zeros([D,D])
    Gamma = Gamma.tolist()
    num11 = transpose(tile(xi2[0],(N[1],1)))
    num22 = tile(xi2[1],(N[0],1))
    num12 = zeros(N)
    for i in arange(N[0]):
        for j in arange(N[1]):
            num12[i][j] = xi[0][i]*xi[1][j]
    denom2 = num11+num22;denom2[0][0]=1
    denom4 = denom2**2
    Gamma[0][0] = C[0]*4*num11/denom2-C[1]*num11**2/denom4
    Gamma[1][0] = Gamma[0][1]=-C[1]*num11*num22/denom4
    Gamma[2][0] = Gamma[0][2]=sqrt(2)*(C[0]*2*num12/denom2-C[1]*transpose(tile(xi[0]**3,(N[1],1)))*tile(xi[1],(N[0],1))/denom4)
    Gamma[1][1] = C[0]*4*num22/denom2-C[1]*tile(xi[1]**4,(N[0],1))/denom4
    Gamma[2][1] = Gamma[1][2]=sqrt(2)*(C[0]*2*num12/denom2-C[1]*transpose(tile(xi[0],(N[1],1)))*tile(xi[1]**3,(N[0],1))/denom4)
    Gamma[2][2] = 2*(C[0]*(num11+num22)/denom2-C[1]*num11*num22/denom4)
    return Gamma

def get_stiff_Man(K,G):
    """Defines a stiffness matrix L in engineering Mandel's notation 
    INPUT = 
        K : Bulk modulus at nodal points; matrix of size N
        G : Shear modulus at nodal points; matrix of size N
    OUTPUT = 
        L : Stiffness matrix of size DxD
    """
    L = array([[K+4/3*G,K-2/3*G,0],[K-2/3*G,K+4/3*G,0],[0,0,2*G]])
    return L

def get_L_Man(K,G):
    """Defines a stiffness matrix L in engineering Mandel's notation  
    INPUT = 
        K : Bulk modulus at nodal points; matrix of size N
        G : Shear modulus at nodal points; matrix of size N
    OUTPUT = 
        L : Stiffness matrix of size DxDxN
    """
    d = K.ndim
    D = d*(d+1)/2
    N = array(K.shape)
    L = zeros(hstack([D,D,N]))
#    L = L.tolist()
    L[0][0] = K+4./3*G;
    L[1][0] = L[0][1] = K-2./3*G;
    L[1][1] = K+4./3*G;
    L[2][0] = L[0][2] = zeros(N);
    L[2][1] = L[1][2] = zeros(N);
    L[2][2] = 2*G;
    return L

def mult_Lep(L,ep,D,DN):
    """Multiplication of matrix L with vector ep 
    INPUT = 
        L : Stiffness matrix of size DxDxN
        ep : Strain vector of size DxN
        D : dimension in engineering notation; D = 3
        DN : the vector consisting the number of components in the dimension D; DN = hstack([D,N])
    OUTPUT = 
        L : Stiffness matrix of size DxDxN
    """
    Lep = zeros(DN)
    for i in arange(D):
        for j in arange(D):
            Lep[i] = Lep[i] + L[i][j]*ep[j]
    return Lep

def mult_GFsig(Gam,Fsig,D,DN):
    """Multiplication of matrix Gam with vector Fsig; in complex field 
    INPUT = 
        Gam : Projection matrix of size DxDxN
        Fsig : Vector of size DxN
        D : dimension in engineering notation; D = 3
        DN : the vector consisting the number of components in the dimension D; DN = hstack([D,N])
    OUTPUT = 
        GFsig : Production of matrix by vector multiplication of size DxDxN
    """
    GFsig = zeros(DN,dtype=complex128)
    for i in arange(D):
        GFsig[i] = Gam[i][0]*Fsig[0]+Gam[i][1]*Fsig[1]+Gam[i][2]*Fsig[2]
    return GFsig

def mult_PL(ep,L,Gam,D,DN):
    """Multiplication of matrix P=F^{-1}*Gam*F with vector ep 
    INPUT = 
        L : Stiffness matrix of size DxDxN
        Gam : Projection matrix in Fourier space of size DxDxN
        ep : Strain vector of size DxN
        D : dimension in engineering notation; D = 3
        DN : the vector consisting the number of components in the dimension D; DN = hstack([D,N])
    OUTPUT = 
        AX : Product of multiplication of size DxDxN
    """
    sig = mult_Lep(L,ep,D,DN)
    Fsig = mult_GFsig(Gam,fft2(sig),D,DN)
    AX = real(ifft2(Fsig))
    return AX

def get_xi(Y,N,d):
    """Return frequencies in Fourier space 
    INPUT =
        Y : the size of periodic unit cell 
        N : ndarray of e.g. stiffness coefficients
        d : dimensionn; d = 2
    OUTPUT =
        xi : the vector of discrete frequencies used in Fourier space
    """
    ind_mean = ceil(N/2.-1)
    xi = []
    for i in arange(d):
        ZN = arange(fix(.5-N[i]/2.),fix(N[i]/2.+1))
        ZNshift = hstack((ZN[ind_mean[i]:N[i]],ZN[0:ind_mean[i]]))
        xi.append(ZNshift/Y[i])
    return xi

def get_xiM(Y,N,d):
    """The auxiliary function enlarging the vector of frequencies used in Fourier space 
    INPUT =
        Y : the size of periodic unit cell 
        N : ndarray of e.g. stiffness coefficients
        d : dimensionn; d = 2
    OUTPUT =
        xiM : the matrix of discrete frequencies used in Fourier space
    """        
    xi = get_xi(Y,N,d)
    xiM = zeros(hstack([d,N]))
    xiM[0] = transpose(tile(xi[0],(N[1],1)))
    xiM[1] = tile(xi[1],(N[0],1))
    return xiM

def get_K0_G0(K,G):
    """Return reference material parameters for projection matrix Gamma 
    INPUT = 
        K : Bulk modulus at nodal points; sizing N
        G : Shear modulus at nodal points; sizing N
    OUTPUT =
        K0 : reference Bulk modulus; constant
        G0 : reference Shear modulus; constant
    """
    K0 = (K.max() + K.min())/2
    G0 = (G.max() + G.min())/2
    return K0,G0
    
def Enu2KG(E,nu):
    """Recalculation of elastic parameters 
    INPUT = 
        E : Young's modulus
        nu : Poisson's ratio
    OUTPUT =
        K : Bulk modulus
        G : Shear modulus 
    """
    K = E/(3*(1-2*nu))
    G = E/(2*(1+nu))
    return K,G

def KG2Enu(K,G):
    """Recalculation of elastic parameters 
    INPUT = 
        K : Bulk modulus
        G : Shear modulus
    OUTPUT = 
        E : Young's modulus
        nu : Poisson's ratio
    """
    E = 9*K*G/(3*K+G)
    nu = (3*K-2*G)/(2*(3*K+G))
    return E,nu