#functions for visco-elastic homogenization
from numpy import *
from numpy.fft import *
from scipy import *
import scipy.optimize
import sys
import dbg
import FFTH_elas
#import matplotlib.pyplot as plt


def homogenization_ve(inputs):
    """FFTH homogenization of linear elasticity
    INPUT = 
        inputs : dictionary consisting of following variables:
            Y : size of periodic unit cell
            tol : tolerance required for convergence
            maxit : maximal number of iterations
            no_time_steps : number of time steps in holding phase
            time_step : time discretization step; \Delta t
            t_L - the duration of loading phase of macroscopic stress
            t_H - the duration of holding phase of macroscopic stress
            E - the matrix from \set{R}^{N_1xN_2} storing the Young's modulus
            nu - the matrix from \set{R}^{N_1xN_2} storing the Poisson's ratio
            eta_M - the matrix from \set{R}^{N_1xN_2} storing the viscoelastic parameter \eta_M
            G_V - the matrix from \set{R}^{N_1xN_2} storing the viscoelastic parameter G_V
            eta_V - the matrix from \set{R}^{N_1xN_2} storing the viscoelastic parameter \eta_V
    OUTPUT =
        outputs:
            L_eff : homogenized (effective) stiffness matrix calculated for E and nu
            K_eff : effective bulk modulus obtained from L_eff
            G_eff : effective shear modulus obtained from L_eff
            etaM_eff : effective visco-elastic parameter etaM
            GV_eff : effective visco-elastic parameter GV
            etaV_eff : effective visco-elastic parameter etaV
            time : time discretization points 
            time_loading : time discretization points in loading phase
            time_holding : time discretization points in holding phase
            strain33_mean - mean of shear strain at time discretization points
            strain33_holding_approx - the approximation of strain33_mean in holding phase with KVM model
            stress33_mean - mean of shear stress at time discretization points
    """
    outputs = dict()
    Dt = inputs['time_step']
    t_L = inputs['t_L']
    t_H = inputs['t_H']
    time_loading = arange(0,t_L,Dt); time_holding = arange(t_L,t_L+t_H+Dt,Dt); time = hstack((time_loading,time_holding));
    
    K = inputs['E']/(3*(1-2*inputs['nu']))
    G = inputs['E']/(2*(1+inputs['nu']))
    etaM = inputs['eta_M']
    GV = inputs['G_V'] 
    etaV = inputs['eta_V']
    
    # t_He solution of linear system wit_H Gve
    beta = exp(-GV*Dt/etaV)
    a = (1/G + Dt/2/etaM + (1-beta)/(2*GV))/2;
    Gve = 1./2/a
    inputs['L'] = FFTH_elas.get_L_Man(K,Gve)
    #possibility to set tol and maxit
    inputs['K0'],inputs['G0'] = FFTH_elas.get_K0_G0(K,G)
    outputs_Gve = FFTH_elas.homogenization(inputs)
    
    d,D,N,DN = FFTH_elas.get_dimensions(inputs['E'])
    #definition of stress loading function
    sig0_max = 1
    sig0_fun = lambda t: (t<t_L)*t/t_L*sig0_max + (t>=t_L)*sig0_max
    
#    plt.figure(1)
#    plt.plot(hstack((time_loading,t_L)),sig0_fun(hstack((time_loading,t_L))),'b')
#    plt.plot(time_holding,sig0_fun(time_holding),'r')
#    plt.plot(time,sig0_fun(time),'xk')
#    plt.ylim(0,1.05)
#    plt.ylabel('Macroscopic stress_12')
#    plt.xlabel('Time')
#    plt.legend( ('Loading phase', 'Holding phase', 'Discretization points'), loc=0)
#    plt.savefig('figures/hom_loading.eps')
    
    # preallocation
    strain33_mean = []; stress33_mean = []; strain33_mean.append(0.); stress33_mean.append(0.)
    invL_eff = linalg.inv(outputs_Gve['L_eff']) #inverse of homogenized stiffness
    
    print('time-level',time[1])
    strain_st = zeros(DN)
    for i in arange(D):
        strain_st = strain_st + sig0_fun(time[1])*invL_eff[2][i]*outputs_Gve['strain'][i]
    stress_st = FFTH_elas.mult_Lep(inputs['L'],strain_st,D,DN)
    
    strain = strain_st.copy()
    stress = stress_st.copy()
    #calculation of partial stresses and strains in t_He constitutive model
    strain_V = (1-beta)*stress[2]/(4*GV)
    strain_M = Dt/etaM*stress[2]/4
    
    strain33_mean.append(mean(strain[2]))
    stress33_mean.append(mean(stress[2]))
    
    #preprocessing
    K0,G0 = FFTH_elas.get_K0_G0(K,Gve)
    Gam = FFTH_elas.get_Gamma(N, d, D, inputs['Y'], K0, G0)
    
    b = zeros(N)
    for ind_t in arange(2,time.size):
        print('time-level',time[ind_t])
        stressP = copy(stress[2])
        strainP_M = strain_M.copy(); strainP_V = strain_V.copy()
        #right-hand side
        b = Dt/4/etaM*stressP + strainP_M + strainP_V*beta + (1-beta)/(4*GV)*stressP;
        Bve = zeros(DN); Bve[2] = -(b)/a
        #calculation of strains
        X = zeros(DN) #initial vector
        B = -mult_P(Bve,Gam,D,DN) #right-hand side of linear system
        strain_Bve, resnorm, kit = FFTH_elas.CG_solver(inputs['L'],B,X,Gam,inputs['tol'],inputs['maxit'])
        stress_Bve = FFTH_elas.mult_Lep(inputs['L'],strain_Bve,D,DN) + Bve
        
        stress_load = zeros(D)
        for i in arange(D):
            stress_load[i] = (i==2)*(sig0_fun(time[ind_t])) - stress_Bve[i].mean()
        
        strain_load = zeros(D)
        for i in arange(D):
            for j in arange(D):
                strain_load[i] = strain_load[i] + invL_eff[i][j]*stress_load[j]
        
        strain = strain_Bve
        for i in arange(D):
            strain = strain + outputs_Gve['strain'][i]*strain_load[i]
        stress = FFTH_elas.mult_Lep(inputs['L'],strain,D,DN) + Bve
        
        #dopocitani jednot_Livych prispevku   
        strain_V = strainP_V*beta + (1-beta)*(stress[2] + stressP)/(4*GV)
        strain_M = Dt/etaM*(stress[2] + stressP)/4 + strainP_M
        
        strain33_mean.append(mean(strain[2]))
        stress33_mean.append(mean(stress[2]))
        
    #the FFT-homogenization with K G
    inputs['L'] = FFTH_elas.get_L_Man(K,G)
    inputs['K0'],inputs['G0'] = FFTH_elas.get_K0_G0(K,G)
    outputs_G = FFTH_elas.homogenization(inputs)
    print('The FFT-homogenization was proceeded.')
    
    ind_holding = arange(time_loading.size,time.size)#indices of holding phase
    strain33_mean = array(strain33_mean)
    
    outputs['G_eff'] = outputs_G['L_eff'][2][2]/2
    outputs['K_eff'] = (outputs_G['L_eff'][1][1]+outputs_G['L_eff'][2][2])/2 - 4/3*outputs['G_eff']
    
    strain_H = lambda p, t: .5*(sig0_max/outputs['G_eff'] + sig0_max/p[1] + (sig0_max*t)/p[0] - (sig0_max*t_L)/(2*p[0]) - (p[2]*sig0_max*(exp((p[1]*t_L)/p[2]) - 1))/(p[1]**2*t_L*exp((p[1]*t)/p[2])))
    eKVM = lambda p, t, y: (strain_H(p,t)-y)**2
    
    p0 = [mean(etaM),mean(GV),mean(etaV)]#initial approximation of fitted parameters
    resultsKVM = scipy.optimize.minpack.leastsq(eKVM, p0, args=(time[ind_holding],strain33_mean[ind_holding]),maxfev=1000)
    parametersKVM_eff = resultsKVM[0]
    
#    plt.figure(2)
#    plt.plot(time[arange(time_loading.size+1)],strain33_mean[arange(time_loading.size+1)],'b')
#    plt.plot(time[ind_holding],strain33_mean[ind_holding],'r')
#    plt.plot(time,strain33_mean,'xk')
#    plt.ylabel('Macroscopis strain_12')
#    plt.xlabel('Time')
#    plt.legend( ('Loading phase', 'Holding phase', 'Discretization points'), loc=0)
#    plt.savefig('figures/hom_strain33.eps')
#    
#    time_plt = linspace(t_L,t_L+t_H)
#    plt.figure(21)
#    plt.plot(time[ind_holding],strain33_mean[ind_holding],'r')
#    plt.plot(time_plt,strain_H(parametersKVM_eff,time_plt),'g')
#    plt.plot(time[ind_holding],strain33_mean[ind_holding],'xk')
#    plt.ylabel('Macroscopic strain_12')
#    plt.xlabel('Time')
#    plt.legend( ('Numerical experiment', 'Approximation wit_H KVM model', 'Discretization points'), loc=0)
#    plt.savefig('figures/hom_strain33_holding.eps')
    
    outputs['etaM_eff'] = resultsKVM[0][0]
    outputs['GV_eff'] = resultsKVM[0][1]
    outputs['etaV_eff'] = resultsKVM[0][2]
    outputs['L_eff'] = outputs_G['L_eff']
    outputs['strain33_mean'] = array(strain33_mean)
    outputs['stress33_mean'] = array(stress33_mean)
    outputs['strain33_holding_approx'] = strain_H(resultsKVM[0],time[ind_holding])
    outputs['time'] = time
    outputs['time_loading'] = time_loading
    outputs['time_holding'] = time_holding
    return outputs

def mult_P(stress,Gam,D,DN):
    Fep = FFTH_elas.mult_GFsig(Gam,fft2(stress),D,DN)
    AX = real(ifft2(Fep));
    return AX
