#file with functions providing analytical homogenization
import string
from numpy import *
#from elas import *
#zeros, arange, prod, ndim, array, float64, flipud, fix, size, mean, abs
#import datetime
#import matplotlib.pyplot as plt
#import io_rw

def Voight_Reuss(K,G,f):
    """Voight-Reuss lower and upper bounds
    INPUT = 
        K : Bulk modulus of individual phases; vector sizing the number of phases
        G : Shear modulus of individual phases; vector sizing the number of phases
        f : Volume ratio of individual phases; vector sizing the number of phases
    OUTPUT = 
        K_V : Voight's Bulk modulus: the upper bound
        G_V : Voight's Shear modulus: the upper bound
        K_R : Reuss's Bulk modulus: the lower bound
        G_R : Reuss's Shear modulus: the lower bound
    """
    K_V = sum(f*K)
    G_V = sum(f*G)
    K_R = 1/sum(f/K)
    G_R = 1/sum(f/G)
    return K_V,G_V,K_R,G_R

def MoriTanaka(K,G,f,Mphase):
    """Mori-Tanaka homogenization
    INPUT = 
        K : Bulk modulus of individual phases; vector sizing the number of phases
        G : Shear modulus of individual phases; vector sizing the number of phases
        f : Volume ratio of individual phases; vector sizing the number of phases
        Mphase : the number determining the position of K,G, and f that corresponds to matrix phase
    OUTPUT = 
        K_hmg : effective (homogenized) Bulk modulus
        G_hmg : effective (homogenized) Shear modulus
    """
    alpha_m = 3. * K[Mphase] / ( 3. * K[Mphase] + 4 * G[Mphase] );
    beta_m  = 6. * ( K[Mphase] + 2. * G[Mphase] ) / ( 5. * ( 3. * K[Mphase] + 4. * G[Mphase] ) );
    
    nom_K_MT = sum(f * K / ( 1. + alpha_m * ( K / K[Mphase] - 1. ) ))
    denom_K_MT = sum(f / ( 1. + alpha_m * ( K / K[Mphase] - 1. ) ))
    nom_G_MT = sum(f * G / ( 1. + beta_m * ( G / G[Mphase] - 1. ) ))
    denom_G_MT = sum(f / ( 1. + beta_m * ( G / G[Mphase] - 1. ) ))

    K_hmg = nom_K_MT / denom_K_MT;
    G_hmg = nom_G_MT / denom_G_MT;
    return K_hmg, G_hmg

def SelfConsistent(K,G,f):
    """Selfconsistent homogenization
    INPUT = 
        K : Bulk modulus of individual phases; vector sizing the number of phases
        G : Shear modulus of individual phases; vector sizing the number of phases
        f : Volume ratio of individual phases; vector sizing the number of phases
    OUTPUT = 
        K_SCS : effective (homogenized) Bulk modulus
        G_SCS : effective (homogenized) Shear modulus
    """
    #first estimation
    K_SCS = mean(K);
    G_SCS = mean(G);
    K_prev = copy(K_SCS);
    G_prev = copy(G_SCS);
    tol = 1e-5
    maxit = 1e3

    res = 1 
    iter = 0
    while res > tol and iter < maxit:
        iter += 1
        alpha_m = 3. * K_SCS / ( 3. * K_SCS + 4 * G_SCS );
        beta_m  = 6. * ( K_SCS + 2. * G_SCS ) / ( 5. * ( 3. * K_SCS + 4. * G_SCS ) );
        nom_K_SCS = sum(f * K / ( 1. + alpha_m * ( K / K_SCS - 1. ) ))
        denom_K_SCS = sum(f / ( 1. + alpha_m * ( K / K_SCS - 1. ) ))
        nom_G_SCS = sum(f * G / ( 1. + beta_m * ( G / G_SCS - 1. ) ))
        denom_G_SCS = sum(f / ( 1. + beta_m * ( G / G_SCS - 1. ) ))
        
        K_SCS = nom_K_SCS / denom_K_SCS;
        G_SCS = nom_G_SCS / denom_G_SCS;
        res = ( (K_SCS-K_prev)**2 + (G_SCS-G_prev)**2 )**.5 
        K_prev = copy(K_SCS);
        G_prev = copy(G_SCS);
    return K_SCS, G_SCS

def HashinShtrikmanWalpole(K,G,f):
    """Hashin-Shtrikman-Walpole upper and lower bounds
    INPUT = 
        K : Bulk modulus of individual phases; vector sizing the number of phases
        G : Shear modulus of individual phases; vector sizing the number of phases
        f : Volume ratio of individual phases; vector sizing the number of phases
        Mphase : the number determining the position of K,G, and f that corresponds to matrix phase
    OUTPUT = 
        K_HSW_upper : the upper bound of Bulk modulus
        G_HSW_upper : the upper bound of Shear modulus
        K_HSW_lower : the lower bound of Bulk modulus
        G_HSW_lower : the lower bound of Shear modulus
    """
    ind_K = K.argsort()
    ind_G = G.argsort()
    G_min = ind_G[0] 
    G_max = ind_G[-1]
    
    K_HSW_lower = 1. / sum( f[ind_K] / ( K[ind_K] + 4./3. * G_min ) ) - 4. / 3. * G_min
    K_HSW_upper = 1. / sum( f[ind_K] / ( K[ind_K] + 4./3. * G_max ) ) - 4. / 3. * G_max
    
    zeta = lambda K, G: G / 6. * ( 9. * K + 8. * G ) / ( K + 2. * G )
#    gamma = lambda f,K,G: sum( f[ind_K] / ( G[ind_K] + zeta(K,G) ) )
    G_HSW_lower = 1. / sum( f[ind_K] / ( G[ind_K] + zeta(K[ind_K[0]],G_min) ) ) - zeta(K[ind_K[0]],G_min)
    G_HSW_upper = 1. / sum( f[ind_K] / ( G[ind_K] + zeta(K[ind_K[-1]],G_max) ) ) - zeta(K[ind_K[-1]],G_max)
    return K_HSW_lower,G_HSW_lower,K_HSW_upper,G_HSW_upper
